from typing import Optional
from ..core.batch import Batch
from ..core.node import Node


class Group(Node):
    """
    Узел группировки данных пакета
    """
    def __init__(
            self,
            class_name: str,
            **kwargs
    ):
        super().__init__(**kwargs)
        self._class_name = class_name

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        idxs = iter(data.metadata.pop('idxs'))
        batch_size = data.metadata.pop('batch_size')
        batch = Batch(data=[{} for _ in range(batch_size)], metadata=data.metadata)

        for obj in data:
            idx = next(idxs)
            batch[idx].setdefault(self._class_name, []).append(obj)

        return batch



import threading
from typing import Any


class Memory:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
                cls._instance._data = {}

        return cls._instance

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, key: Any) -> Any:
        return self._data[key]

    def __setitem__(self, key: Any, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: Any) -> None:
        del self._data[key]

    def __contains__(self, key: Any) -> bool:
        return key in self._data

    def __str__(self) -> str:
        return str(self._data)



from .core import Batch, NodeType, BatchStatus, Node, Watchdog
from .nodes import Group, NodeFastAPI, SyncORT, Ungroup

__all__ = [
    "Batch",
    "NodeType",
    "BatchStatus",
    "Node",
    "Watchdog",
    "Group",
    "NodeFastAPI",
    "SyncORT",
    "Ungroup"
]



from enum import Enum, auto


class NodeType(Enum):
    THREAD = auto()
    PROCESS = auto()


class NodeStatus(Enum):
    PENDING = auto()
    ACTIVE = auto()
    COMPLETED = auto()
    TERMINATED = auto()


class BatchStatus(Enum):
    OK = auto()
    LAST = auto()
    ERROR = auto()



from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Tuple, Any, Union
import onnxruntime
import numpy as np
import cv2

from ..core.batch import Batch
from ..core.node import Node


def resize(
        image: np.ndarray,
        size: Tuple[int, int],
        pad_value: int = 0
) -> np.ndarray:
    """
    Алгоритм изменения размеров изображений
    """
    h, w, c = image.shape

    if (h, w) == (size[0], size[1]):
        return image

    ratio = min(size[0] / h, size[1] / w)
    h_new, w_new = round(h * ratio), round(w * ratio)

    image_resized = cv2.resize(image, (0, 0), fx=ratio, fy=ratio, interpolation=cv2.INTER_AREA)
    image_padded = np.full((size[0], size[1], c), fill_value=pad_value, dtype=image.dtype)

    h_pos, w_pos = (size[0] - h_new) // 2, (size[1] - w_new) // 2
    image_padded[h_pos: h_pos + h_new, w_pos: w_pos + w_new] = image_resized

    return image_padded


class SyncORT(Node, ABC):
    def __init__(
            self,
            model_path: str,
            providers: List[Union[str, Tuple[str, Dict]]],
            pad_value: int = 0,
            data_key: Optional[str] = None,
            half_precision: bool = False,
            **kwargs
    ):
        super().__init__(**kwargs)
        self._session = onnxruntime.InferenceSession(path_or_bytes=model_path, providers=providers)
        self._pad_value = pad_value
        self._dtype = np.float16 if half_precision else np.float32

        net_input = self._session.get_inputs()[0]
        self._input_name = net_input.name
        self._batch_size = net_input.shape[0]

        if isinstance(self._batch_size, str):
            self._batch_size = 32

        self._size = (net_input.shape[-1], net_input.shape[-2])
        self._batch = np.ones((self._batch_size, net_input.shape[1], self._size[1], self._size[0]), dtype=self._dtype)
        self._data_key = data_key

    def _preprocessing(self, image: np.ndarray) -> Tuple[np.ndarray, Dict]:
        meta = dict(original_shape=image.shape)
        image = resize(image, self._size, self._pad_value)
        meta.update(current_shape=image.shape)
        image = self.preprocessing(image)

        return image, meta

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        images = []
        metas = []

        for obj in data:
            image, meta = self._preprocessing(obj[self._data_key] if isinstance(obj, dict) else obj)
            images.append(image)
            metas.append(meta)

        batch = Batch(metadata=data.metadata)

        for i in range(0, len(images), self._batch_size):
            for batch_idx, image in enumerate(images[i: i + self._batch_size]):
                self._batch[batch_idx] = image

            inference = self._session.run(None, {self._input_name: self._batch})

            for *output, meta in zip(*inference, metas[i: i + self._batch_size]):
                batch.append(self.postprocessing(output, meta))

        return batch

    @abstractmethod
    def preprocessing(self, image: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def postprocessing(self, output: List, meta: Dict) -> Any:
        pass



from typing import Optional, List
# from threading import Thread
# from fastapi import FastAPI, APIRouter, Request
# from fastapi.templating import Jinja2Templates
# from fastapi.staticfiles import StaticFiles
# from fastapi.responses import HTMLResponse
# from statistics import mean, pstdev
# from datetime import datetime
# from os.path import join, dirname, abspath
# import uvicorn
import logging
import signal
import time
import os

from .batch import Batch
from .enums import NodeType, NodeStatus, BatchStatus
from .node import Node
from .queue_wrapper import QueueWrapper


class Monitoring(Node):
    def __init__(self, **kwargs):
        super().__init__(node_type=NodeType.PROCESS, timeout=10.0, **kwargs)

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        if time.time() - self.shared_memory['time'] >= 60.0:
            logging.info(f'{self.name}: os.kill(1, signal.SIGTERM)')
            os.kill(1, signal.SIGTERM)


class Watchdog(Node):
    def __init__(
            self,
            pipeline: List[Node],
            *,
            verbose_metrics: bool = False,
            metrics_host: str = '0.0.0.0',
            metrics_port: int = 8080,
            **kwargs
    ):
        super().__init__(timeout=1.0, **kwargs)
        logging.getLogger().setLevel(logging.INFO)
        self._pipeline = pipeline
        self._build_pipeline()
        self.start()
        Monitoring().start()

        # if verbose_metrics:
        #     self._templates = Jinja2Templates(directory=join(dirname(abspath(__file__)), 'templates'))
        #     app = FastAPI()
        #     app.mount('/static', StaticFiles(directory=join(dirname(abspath(__file__)), 'static'), html=True), 'static')
        #     router = APIRouter()
        #     router.add_api_route('/metrics_html', self._metrics_html, methods=['GET'], response_class=HTMLResponse)
        #     router.add_api_route('/metrics_json', self._metrics_json, methods=['GET'])
        #     app.include_router(router)
        #     Thread(target=uvicorn.run, kwargs=dict(app=app, host=metrics_host, port=metrics_port)).start()

    # def _metrics_html(self, request: Request):
    #     return self._templates.TemplateResponse('home.html', dict(request=request))
    #
    # def _metrics_json(self):
    #     metrics = []
    #
    #     for node in self._pipeline:
    #         metrics.append(dict(name=node.name,
    #                             input=node.metrics['input'],
    #                             output=node.metrics['output'],
    #                             duration_mean_ms=f"{mean(node.metrics['duration']) * 1000:.2f}",
    #                             duration_std_ms=f"{pstdev(node.metrics['duration']) * 1000:.2f}"))
    #
    #     return dict(current_time=datetime.now().strftime('%Y-%m-%d %H:%M:%S'), metrics=metrics)

    def _build_pipeline(self):
        queues = {}

        for node in self._pipeline:
            for queue in node.input + node.output:
                if queue in queues:
                    continue

                queues[queue] = QueueWrapper(name=queue, type=node.node_type, maxsize=16)

        for node in self._pipeline:
            for queue in node.input:
                node.input_queues.append(queues[queue])

            for queue in node.output:
                node.output_queues.append(queues[queue])

            node.start()

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        self.shared_memory['time'] = time.time()

        if all(node.is_completed for node in self._pipeline):
            logging.info(f'{self.name}: All nodes finished successfully')

            return Batch(status=BatchStatus.LAST)

        elif any(node.is_terminated for node in self._pipeline):
            logging.info(f'{self.name}: At least one of the nodes has terminated')

            for node in self._pipeline:
                node.drain()
                logging.info(f'{node.name}: Draining')

            return Batch(status=BatchStatus.LAST)



from typing import Optional, Dict, Any, List
from .enums import BatchStatus


class Batch:
    def __init__(
            self,
            data: Optional[List[Any]] = None,
            metadata: Optional[Dict] = None,
            status: BatchStatus = BatchStatus.OK,
            error: Optional[str] = None
    ):
        self._data = data.copy() if data else []
        self._metadata = metadata.copy() if metadata else {}
        self._status = status
        self._error = error

    @property
    def metadata(self) -> Optional[Dict]:
        return self._metadata

    @property
    def is_ok(self) -> bool:
        return self._status == BatchStatus.OK

    @property
    def is_last(self) -> bool:
        return self._status == BatchStatus.LAST

    @property
    def is_error(self) -> bool:
        return self._status == BatchStatus.ERROR

    @property
    def error(self) -> Optional[str]:
        return self._error

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, index: int) -> Any:
        return self._data[index]

    def __setitem__(self, index: int, value: Any) -> None:
        self._data[index] = value

    def __delitem__(self, index: int) -> None:
        del self._data[index]

    def __str__(self) -> str:
        return str(self._data)

    def append(self, item: Any) -> None:
        self._data.append(item)

    def extend(self, iterable: List[Any]) -> None:
        self._data.extend(iterable)



import queue
import multiprocessing
from typing import Any
from .enums import NodeType


class QueueWrapper:
    def __init__(
            self,
            name: str,
            type: NodeType = NodeType.THREAD,
            maxsize: int = 0
    ):
        self._name = name
        self._queue = queue.Queue(maxsize=maxsize) if type == NodeType.THREAD else multiprocessing.Queue(maxsize=maxsize)

    @property
    def name(self) -> str:
        return self._name

    def get(self, *args, **kwargs) -> Any:
        return self._queue.get(*args, **kwargs)

    def put(self, *args, **kwargs) -> None:
        return self._queue.put(*args, **kwargs)

    def empty(self) -> bool:
        return self._queue.empty()

    def size(self) -> int:
        return self._queue.qsize()



import time
import logging
from abc import ABC, abstractmethod
from collections import deque
from threading import Thread
from multiprocessing import Process, Manager
from multiprocessing.managers import DictProxy
from typing import Optional, List, Dict, Union, Any

from .batch import Batch
from .enums import NodeType, NodeStatus, BatchStatus
from .memory import Memory
from .queue_wrapper import QueueWrapper


class Node(ABC):
    _shared_memory = Manager().dict()

    def __init__(
            self,
            node_type: NodeType = NodeType.THREAD,
            input: Optional[Union[str, List[str]]] = None,
            output: Optional[Union[str, List[str]]] = None,
            timeout: float = 0.0
    ):
        self._name = self.__class__.__name__
        self._node_type = node_type
        self._status = NodeStatus.PENDING

        self._input = [] if input is None else [input] if isinstance(input, str) else input
        self._output = [] if output is None else [output] if isinstance(output, str) else output
        self._input_queues = []
        self._output_queues = []

        self._memory = Memory()
        # self._metrics = dict(duration=deque([0], maxlen=100), input=0, output=0)
        self._timeout = timeout

    @property
    def name(self) -> str:
        return self._name

    @property
    def node_type(self) -> NodeType:
        return self._node_type

    @property
    def is_pending(self) -> bool:
        return self._status == NodeStatus.PENDING

    @property
    def is_active(self) -> bool:
        return self._status == NodeStatus.ACTIVE

    @property
    def is_completed(self) -> bool:
        return self._status == NodeStatus.COMPLETED

    @property
    def is_terminated(self) -> bool:
        return self._status == NodeStatus.TERMINATED

    @property
    def input(self) -> List[str]:
        return self._input

    @property
    def output(self) -> List[str]:
        return self._output

    @property
    def input_queues(self) -> List[QueueWrapper]:
        return self._input_queues

    @property
    def output_queues(self) -> List[QueueWrapper]:
        return self._output_queues

    @property
    def memory(self) -> Memory:
        return self._memory

    @property
    def shared_memory(self) -> DictProxy:
        return self._shared_memory

    # @property
    # def metrics(self) -> Dict:
    #     return self._metrics

    def start(self):
        if self._status != NodeStatus.PENDING:
            raise RuntimeError

        self._status = NodeStatus.ACTIVE

        if self._node_type == NodeType.THREAD:
            Thread(target=self._run, name=self._name).start()
        elif self._node_type == NodeType.PROCESS:
            Process(target=self._run, name=self._name).start()

    def post_init(self):
        pass

    def release(self):
        pass

    def _run(self):
        self.post_init()

        while self._status == NodeStatus.ACTIVE:
            input = self._get()

            if isinstance(input, Batch):
                if input.is_last:
                    self._put(input)
                    self._status = NodeStatus.COMPLETED
                    break
                elif input.is_error:
                    logging.error(f'{self.name}: {input.error}')
                    self._status = NodeStatus.TERMINATED
                    break

            output = self._step(input)

            if isinstance(output, Batch):
                if output.is_ok:
                    self._put(output)
                elif output.is_last:
                    self._put(output)
                    self._status = NodeStatus.COMPLETED
                    break
                elif output.is_error:
                    logging.error(f'{self.name}: {output.error}')
                    self._status = NodeStatus.TERMINATED
                    break

            time.sleep(self._timeout)

        self.release()

    def _get(self) -> Optional[Batch]:
        if not self._input_queues:
            return None

        if len(self._input_queues) == 1:
            return self._input_queues[0].get()

        batches = [queue.get() for queue in self._input_queues]

        if len(set(len(batch) for batch in batches)) != 1:
            return Batch(status=BatchStatus.ERROR, error='Length batches cannot be different')

        if all(batch.is_ok for batch in batches):
            status = BatchStatus.OK
        elif all(batch.is_end for batch in batches):
            status = BatchStatus.LAST
        else:
            return Batch(status=BatchStatus.ERROR, error='Batches cannot have different status')

        metadata = {}

        for batch in batches:
            metadata.update(batch.metadata)

        data = [
            {
                queue.name: batch[idx]
                for queue, batch in zip(self._input_queues, batches)
            }
            for idx in range(len(batches[0]))
        ]

        return Batch(data=data, metadata=metadata, status=status)

    def _step(self, input: Optional[Batch]) -> Optional[Batch]:
        try:
            # monotonic = time.monotonic()
            output = self.processing(input)
            # self._metrics['duration'].append(time.monotonic() - monotonic)
            # self._metrics['input'] += len(input) if isinstance(input, Batch) else 0
            # self._metrics['output'] += len(output) if isinstance(output, Batch) else 0
        except Exception as e:
            output = Batch(status=BatchStatus.ERROR, error=f'During processing raise exception {e.__class__} {e}')

        return output

    def _put(self, output: Batch):
        if not self._output_queues:
            return

        for queue in self._output_queues:
            queue.put(output)

    def drain(self):
        self._status = NodeStatus.TERMINATED

        for queue in self._input_queues + self._output_queues:
            while not queue.empty():
                queue.get()

    @abstractmethod
    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        pass
